/*
 * Copyright 2000 by Hans Reiser, licensing governed by reiserfs/README
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <asm/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/vfs.h>
#include <netinet/in.h>
//#include <linux/byteorder/little_endian.h>

//#include "inode.h"
#include "io.h"
#include "misc.h"
#include "reiserfs_fs.h"
#include "../version.h"



int g_new_internals;


#define print_usage_and_exit() die ("Usage: %s [-b block-to-print][-idc] device\n\
-i Causes to print all items of a leaf\n\
-d                 content of directory items\n\
-c                 content of direct items\n\
-m                 bitmap blocks\n", argv[0]);



struct reiserfs_fsstat {
  int nr_internals;
  int nr_leaves;
  int nr_files;
  int nr_directories;
  int nr_unformatted;
} g_stat_info;
int g_comp_number = 0;

/*
 *  options
 */
int opt_print_regular_file_content = 0;/* -c */
int opt_print_details = 0;	/* -d */
int opt_print_leaf_items = 0;		/* -i */
int opt_print_objectid_map = 0;	/* -o */
int opt_print_block_map = 0;		/* -m */
/* when you want print one block specify -b # */
int opt_block_to_print = -1;
int opt_pack = 0;	/* -P will produce output that should be |gzip -c > whatever.gz */
			/* -p will calculate number of bytes needed to transfer the partition */
int opt_print_journal;
int opt_pack_all = 0;

struct super_block g_sb;



int print_mode (void)
{
    int mode = 0;

    if (opt_print_leaf_items == 1)
	mode |= PRINT_LEAF_ITEMS;
    if (opt_print_details == 1)
	mode |= (PRINT_LEAF_ITEMS | PRINT_ITEM_DETAILS);
    if (opt_print_regular_file_content == 1)
	mode |= (PRINT_LEAF_ITEMS | PRINT_DIRECT_ITEMS);
    return mode;
}


void print_disk_tree (int block_nr)
{
    struct buffer_head * bh;

    bh = bread (g_sb.s_dev, block_nr, g_sb.s_blocksize);

    if (is_internal_node (bh)) {
	int i;

	g_stat_info.nr_internals ++;
	print_block (bh, print_mode (), -1, -1);

#if 1
	// XXX JDM I think this is cleaner..
        for (i = 0; i <= B_NR_ITEMS (bh); i ++)
            print_disk_tree( child_block_number(bh,i) );
#else
	{
	struct disk_child * dc;
        dc = B_N_CHILD (bh, 0);
        for (i = 0; i <= B_NR_ITEMS (bh); i ++, dc ++)
            print_disk_tree ( le32_to_cpu( dc->dc_block_number ));
	}
 
#endif
      
    } else if (is_leaf_node (bh)) {
		g_stat_info.nr_leaves++;
		// XXX JDM This causes gcc to bus error w/ -O2 on ppc (???)
		print_block( bh, print_mode(), -1, -1 );
    } else {
		print_block (bh, print_mode (), -1, -1);
		die ("print_disk_tree: bad block type");
    }
    brelse (bh);
}



void print_one_block (int block)
{
    struct buffer_head * bh;
    
    if (test_bit (block % (g_sb.s_blocksize * 8), 
		  SB_AP_BITMAP (&g_sb)[block / (g_sb.s_blocksize * 8)]->b_data))
	fprintf (stderr, "%d is used in true bitmap\n", block);
    else
	fprintf (stderr, "%d is free in true bitmap\n", block);
    
    bh = bread (g_sb.s_dev, block, g_sb.s_blocksize);
    if (!bh) {
	printf ("print_one_block: bread fialed\n");
	return;
    }

    if (opt_pack) {
	pack_one_block (&g_sb, bh);
	brelse (bh);
	return;
    }

    if (who_is_this (bh->b_data, g_sb.s_blocksize) != THE_UNKNOWN)
	print_block (bh, PRINT_LEAF_ITEMS | PRINT_ITEM_DETAILS | (opt_print_regular_file_content == 1 ? PRINT_DIRECT_ITEMS : 0), -1, -1);
    else
	printf ("Looks like unformatted\n");
    brelse (bh);
    return;
}


static char * parse_options (int argc, char * argv [])
{
    int c;
    char * tmp;
  
    while ((c = getopt (argc, argv, "b:icdmoMpPaAj")) != EOF) {
	switch (c) {
	case 'b':	/* print a single node */
	    opt_block_to_print = strtol (optarg, &tmp, 0);
	    if (*tmp)
		die ("parse_options: bad block size");
	    break;

	case 'p':	/* calculate number of bytes, that need to be transfered */
	    opt_pack = 'c'; break;
	case 'P':	/* dump a partition */
	    opt_pack = 'p'; break;
	case 'a':
	    opt_pack_all = 'c'; break;
	case 'A':
	    opt_pack_all = 'p'; break;

	case 'i':	/* print items of a leaf */
	    opt_print_leaf_items = 1; break;

	case 'd':	/* print directories */
	    opt_print_details = 1; break;

	case 'c':	/* print contents of a regular file */
	    opt_print_regular_file_content = 1; break;

	case 'o':	/* print a objectid map */
	    opt_print_objectid_map = 1; break;

	case 'm':	/* print a block map */
	    opt_print_block_map = 1;  break;
	case 'M':	/* print a block map with details */
	    opt_print_block_map = 2;  break;
	case 'j':
	    opt_print_journal = 1; break; /* print transactions */
	}
    }
    if (optind != argc - 1)
	/* only one non-option argument is permitted */
	print_usage_and_exit();
  
    return argv[optind];
}


#if 0

/* journal has permanent location (currently) (after first bitmap
   block and constant size (JOURNAL_BLOCK_COUNT + 1) */
int journal_block (struct super_block * s, __u32 block)
{
/*
    if (block > SB_AP_BITMAP (s)[0]->b_blocknr &&
	block < SB_AP_BITMAP (s)[0]->b_blocknr + JOURNAL_BLOCK_COUNT + 1)
	return 1;
*/
    if (block >= SB_JOURNAL_BLOCK (s) &&
	block <= SB_JOURNAL_BLOCK (s) + rs_journal_size (s->u.reiserfs_sb.s_rs) + 1)
	return 1;
    return 0;
}


int data_block (struct super_block * s, __u32 block)
{
    int i;

    if (block == REISERFS_DISK_OFFSET_IN_BYTES / s->s_blocksize)
      /* super block, not data block */
      return 0;

    for (i = 0; i < SB_BMAP_NR (s); i ++)
	if (block == SB_AP_BITMAP (s)[i]->b_blocknr)
	    /* bitmap block, not data block */
	    return 0;

    if (journal_block (s, block))
	return 0;

    return 1;
}

#endif



/* print all valid transactions and found dec blocks */
static void print_journal (struct super_block * s)
{
    struct buffer_head * d_bh, * c_bh;
    struct reiserfs_journal_desc * desc ;
    struct reiserfs_journal_commit *commit ;
    int end_journal;
    int start_journal;
    int i, j;
    int first_desc_block = 0;
    int wrapped = 0;
    int valid_transactions = 0;

    start_journal = SB_JOURNAL_BLOCK (s);
    end_journal = start_journal + JOURNAL_BLOCK_COUNT;
    printf ("Start scanning from %d\n", start_journal);

    d_bh = 0;
    desc = 0;
    for (i = start_journal; i < end_journal; i ++) {
	d_bh = bread (s->s_dev, i, s->s_blocksize);
	if (who_is_this (d_bh->b_data, d_bh->b_size) == THE_JDESC) {
	    int commit_block;

	    if (first_desc_block == 0)
		/* store where first desc block found */
		first_desc_block = i;

	    print_block (d_bh); /* reiserfs_journal_desc structure will be printed */
	    desc = (struct reiserfs_journal_desc *)(d_bh->b_data);

	    commit_block = d_bh->b_blocknr + desc->j_len + 1;
	    if (commit_block >= end_journal) {
		printf ("-- wrapped?");
		wrapped = 1;
		break;
	    }

	    c_bh = bread (s->s_dev, commit_block, s->s_blocksize);
	    commit = bh_commit (c_bh);
	    if (does_desc_match_commit (s, desc, commit)) {
		printf ("commit block %d (trans_id %ld, j_len %ld) does not match\n", commit_block,
			commit->j_trans_id, commit->j_len);
		brelse (c_bh) ;
		brelse (d_bh);
		continue;
	    }

	    valid_transactions ++;
	    printf ("(commit block %d) - logged blocks (", commit_block);
	    for (j = 0; j < desc->j_len; j ++) {
		if (j < JOURNAL_TRANS_HALF) {
		    printf (" %ld", desc->j_realblock[j]);
		} else {
		    printf (" %ld", commit->j_realblock[i - JOURNAL_TRANS_HALF]);
		}
	    }
	    printf ("\n");
	    i += desc->j_len + 1;
	    brelse (c_bh);
	}
	brelse (d_bh);
    }
    
    if (wrapped) {
	c_bh = bread (s->s_dev, first_desc_block - 1, s->s_blocksize);
	commit = bh_commit (c_bh);
	if (does_desc_match_commit (s, desc, commit)) {
	    printf ("No! commit block %d (trans_id %ld, j_len %ld) does not match\n", first_desc_block - 1,
		    commit->j_trans_id, commit->j_len);
	} else {
	    printf ("Yes! (commit block %d) - logged blocks (\n", first_desc_block - 1);
	    for (j = 0; j < desc->j_len; j ++) {
		if (j < JOURNAL_TRANS_HALF) {
		    printf (" %ld", desc->j_realblock[j]);
		} else {
		    printf (" %ld", commit->j_realblock[i - JOURNAL_TRANS_HALF]);
		}
	    }
	    printf ("\n");
	}
	brelse (c_bh) ;
	brelse (d_bh);
    }

    printf ("%d valid transactions found\n", valid_transactions);

    {
	struct buffer_head * bh;
	struct reiserfs_journal_header * j_head;

	bh = bread (s->s_dev, SB_JOURNAL_BLOCK (s) + rs_journal_size (s->s_rs),
		    s->s_blocksize);
	j_head = (struct reiserfs_journal_header *)(bh->b_data);

	printf ("#######################\nJournal header:\n"
		"j_last_flush_trans_id %ld\n"
		"j_first_unflushed_offset %ld\n"
		"j_mount_id %ld\n", j_head->j_last_flush_trans_id, j_head->j_first_unflushed_offset,
		j_head->j_mount_id);
	brelse (bh);
    }
}



int main (int argc, char * argv[])
{
    char * file_name;
    int dev, i;

#ifdef USE_ENDIAN_CONVERSION_MACROS
    printf( "Using endian conversion macros...\n" );
#endif
    printf( "sizeof( struct stat_data ) = %d\n", sizeof( struct stat_data ) );
    printf( "sizeof( struct stat_data_v1 ) = %d\n", sizeof( struct stat_data_v1) );                                                                            

    print_banner ("debugreiserfs");
 

    file_name = parse_options (argc, argv);


    dev = open (file_name, O_RDONLY);
    if (dev == -1)
	die ("dumpreiserfs: Can not open device %s: %s\n", file_name, strerror (errno));
    g_sb.s_dev = dev;

    if (uread_super_block (&g_sb))
	die ("dumpreiserfs: no reiserfs found on %s", file_name);
    if (uread_bitmaps (&g_sb))
	die ("dumpreiserfs: read_bitmap failed");

    if (opt_block_to_print != -1) {
      // pack or print node
	print_one_block (opt_block_to_print);
	goto end;
    }

    if (opt_pack || opt_pack_all) {
	pack_partition (&g_sb);
    } else {
	/* dump file system to stdout */

	print_block (SB_BUFFER_WITH_SB (&g_sb));

	if (opt_print_journal)
	    print_journal (&g_sb);
    
	if (opt_print_objectid_map == 1)
	    print_objectid_map (&g_sb);
    
	if (opt_print_block_map) {
	    print_bmap (&g_sb, opt_print_block_map == 1 ? 1 : 0);
	}

	if (opt_print_regular_file_content || opt_print_details ||
	    opt_print_leaf_items) {
	    print_disk_tree (get_root_block (&g_sb));

	    /* print the statistic */
	    printf ("File system uses %d internal + %d leaves + %d unformatted nodes = %d blocks\n", 
		    g_stat_info.nr_internals, g_stat_info.nr_leaves, g_stat_info.nr_unformatted, 
		    g_stat_info.nr_internals + g_stat_info.nr_leaves + g_stat_info.nr_unformatted);
	}
    }


 end:
    {
	int bmap_num;

	bmap_num = get_bmap_num (&g_sb);
	
	/* brelse bitmaps */
	if (SB_AP_BITMAP (&g_sb)) {
	    for (i = 0; i < bmap_num; i ++) {
		brelse (SB_AP_BITMAP (&g_sb)[i]);
	    }
	    freemem (SB_AP_BITMAP (&g_sb));
	}
    }

    /* brelse buffer containing super block */
    brelse (SB_BUFFER_WITH_SB (&g_sb));

    check_and_free_buffer_mem ();

    return 0;
}
/* end of main */





