/*
 * Copyright 2000 by Hans Reiser, licensing governed by reiserfs/README
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <malloc.h>
#include <sys/types.h>
#include <asm/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "../include/io.h"
#include "../include/misc.h"
#include "../include/reiserfs_fs.h"


// the leaf is stored in compact form: 
// start magic number
// block number __u32
// item number __u16
// struct packed_item
// ..
// end magic number

#define LEAF_START_MAGIC 0xa5
#define LEAF_END_MAGIC 0x5a
#define FULL_BLOCK_START_MAGIC 0xb6
#define FULL_BLOCK_END_MAGIC 0x6b
#define UNFORMATTED_BITMAP_START_MAGIC 0xc7
#define UNFORMATTED_BITMAP_END_MAGIC 0x7c
#define END_MAGIC 0x8d

#define DIR_ID 1
#define OBJECT_ID 2
#define OFFSET32 4
#define OFFSET64 8
#define ENTRY_COUNT 16 // shows whether ih_free_space/ih_entry_count is stored
#define NEW_FORMAT 32  // 0 here means - old format, 1 - new format
#define ITEM_BODY 64
#define WHOLE_INDIRECT 128
#define INDIRECT_ITEM 256
#define DIRENTRY_ITEM 512

struct packed_item {
    __u16 length; // length of the area we store item in
    __u16 mask; // what is stored: dirid, objectid, 32 bit offset or 64 bit offset, type
    __u16 item_len;
};


#define fread16(pv) fread (pv, sizeof (__u16), 1, stdin)
#define fread32(pv) fread (pv, sizeof (__u32), 1, stdin)
#define fread64(pv) fread (pv, sizeof (__u64), 1, stdin)

#define fwrite16(pv) fwrite (pv, sizeof (__u16), 1, stdout)
#define fwrite32(pv) fwrite (pv, sizeof (__u32), 1, stdout)
#define fwrite64(pv) fwrite (pv, sizeof (__u64), 1, stdout)
