/*
 * Copyright 2000 by Hans Reiser, licensing governed by reiserfs/README
 */

#include "fsck.h"
#include <math.h>

#if 0 /* original function */
//
// FIXME: we are going to turn reuse of objectids back
//
__u32 get_unused_objectid (struct super_block * s)
{
    __u32 * map, objectid;

    map = (__u32 *)((char *)SB_DISK_SUPER_BLOCK (s) + (sb_size(s)));
    objectid = map[1];
    if (objectid == TYPE_INDIRECT)
	die ("get_unused_objectid: no more objectids");

    map[1] ++;
    return objectid;
}
#endif

#if 0
/* stolen from reiserfs/objectid.c  --clm */
__u32 get_unused_objectid (struct super_block * s)
{
    __u32 unused_objectid;
    struct reiserfs_super_block * rs;
    __u32 * objectid_map;


    rs = s->u.reiserfs_sb.s_rs;

    /* The objectid map follows the superblock. */
    objectid_map = (__u32 *)((char *)rs + (sb_size(s)));
 
    unused_objectid = objectid_map[1]; /* commented more below */
    if (unused_objectid == TYPE_INDIRECT) {
	printf ("REISERFS: get_objectid: no more object ids\n");
	return 0;
    }

    /* This incrementation allocates the first unused objectid. That is to say, 
       the first entry on the objectid map is the first unused objectid, and 
       by incrementing it we use it.  See below where we check to see if we 
       eliminated a sequence of unused objectids.... */

    objectid_map[1] ++;

    /* Now we check to see if we eliminated the last remaining member of
       the first even sequence (and can eliminate the sequence by
       eliminating its last objectid from oids), and can collapse the
       first two odd sequences into one sequence.  If so, then the net
       result is to eliminate a pair of objectids from oids.  We do this
       by shifting the entire map to the left. */
    if (rs_objectid_map_size (rs) > 2 && objectid_map[1] == objectid_map[2]) {
	memmove (objectid_map + 1, objectid_map + 3, (rs_objectid_map_size (rs) - 3) * sizeof(__u32));
	set_objectid_map_size (rs, rs_objectid_map_size (rs) - 2);
    }

    return unused_objectid;
}

int is_objectid_used (struct super_block * s, __u32 objectid)
{
    __u32 * objectid_map;
    int i = 0;

    objectid_map = (__u32 *)((char *)SB_DISK_SUPER_BLOCK (s) + (sb_size(s)));
  
    while (i < SB_OBJECTID_MAP_SIZE (s)) {
	if (objectid == objectid_map[i]) {
	    return 1;      /* objectid is used */
	}
	
	if (objectid > objectid_map[i] && objectid < objectid_map[i+1]) {
	    return 1;	/* objectid is used */
	}
	
	if (objectid < objectid_map[i])
	    break;

	i += 2;
    }
    
    /* objectid is free */
    return 0;
}


/* we mark objectid as used. Additionally, some unused objectids can
   become used. It is ok. What is unacceptable, it is when used
   objectids are marked as unused */
void mark_objectid_as_used (struct super_block * s, __u32 objectid)
{
    int i;
    __u32 * objectid_map;
    
    if (is_objectid_used (s, objectid) == 1) {
	/*print_objectid_map (s);*/
	/*printf ("mark_objectid_as_used: objectid %lu is used", objectid);*/
	return;
    }

    objectid_map = (__u32 *)((char *)SB_DISK_SUPER_BLOCK (s) + (sb_size(s)));

    for (i = 0; i < SB_OBJECTID_MAP_SIZE (s); i += 2) {
	if (objectid >= le32_to_cpu (objectid_map [i]) && 
	    objectid < le32_to_cpu (objectid_map [i + 1]))
	    /* it is used */
	    return;

	if (objectid + 1 == le32_to_cpu (objectid_map[i])) {
	    /* size of objectid map is the same */
	    objectid_map[i] = cpu_to_le32 (objectid);
	    return;
	}

	if (objectid == le32_to_cpu (objectid_map[i + 1])) {
	    /* size of objectid map is decreased */
	    objectid_map[i + 1] = cpu_to_le32 (le32_to_cpu (objectid_map[i + 1]) + 1);
	    if (i + 2 < SB_OBJECTID_MAP_SIZE (s)) {
		if (objectid_map[i + 1] == objectid_map[i + 2]) {
		    memmove (objectid_map + i + 1, objectid_map + i + 1 + 2, 
			     (SB_OBJECTID_MAP_SIZE (s) - (i + 2 + 2 - 1)) * sizeof (__u32));
		    
		    set_objectid_map_size (s->u.reiserfs_sb.s_rs, SB_OBJECTID_MAP_SIZE (s) - 2);
		    //SB_DISK_SUPER_BLOCK (s)->s_oid_cursize -= 2;
		}
	    }
	    return;
	}
	
	if (objectid < le32_to_cpu (objectid_map[i])) {
	    /* size of objectid map must be increased */
	    if (SB_OBJECTID_MAP_SIZE (s) == SB_OBJECTID_MAP_MAXSIZE (s)) {
		/* here all objectids between objectid and objectid_map[i] get used */
		objectid_map[i] = cpu_to_le32 (objectid);
		return;
	    } else {
		memmove (objectid_map + i + 2, objectid_map + i, (SB_OBJECTID_MAP_SIZE (s) - i) * sizeof (__u32));
		set_objectid_map_size (s->u.reiserfs_sb.s_rs, SB_OBJECTID_MAP_SIZE (s) + 2);
	    }
	    
	    objectid_map[i] = cpu_to_le32 (objectid);
	    objectid_map[i+1] = cpu_to_le32 (objectid + 1);
	    return;
	}
	
    }
    
    /* write out of current objectid map, if we have space */
    if (i < SB_OBJECTID_MAP_MAXSIZE (s)) {
	objectid_map[i] = cpu_to_le32 (objectid);
	objectid_map[i + 1] = cpu_to_le32 (objectid + 1);
	set_objectid_map_size (s->u.reiserfs_sb.s_rs, SB_OBJECTID_MAP_SIZE (s) + 2);
    } else if (i == SB_OBJECTID_MAP_MAXSIZE (s)) { //SB_DISK_SUPER_BLOCK (&g_sb)->s_oid_maxsize) {
	objectid_map[i - 1] = cpu_to_le32 (objectid + 1);
    } else
	die ("mark_objectid_as_used: objectid map corrupted");
    
    return;
}
#endif

/* size of 1 piece of map */
#define MAP_SIZE   1024 /* must be n * 2 * sizeof(__u32) */

struct{
    __u32 * m_begin; // pointer to map area
    __u32   m_used;  // used places in area <= m_count * MAP_SIZE / sizeof (__u32)
    __u32   m_count; // count of MAPs
}id_map;			

void init_id_map()
{
    id_map.m_begin = NULL;
    id_map.m_used = 0;
    id_map.m_count = 0;
    mark_id_as_used(1);
}

/* increase area on MAP_SIZE bytes */
static void grow_id_map()
{
    id_map.m_begin = expandmem(id_map.m_begin, id_map.m_count * MAP_SIZE, MAP_SIZE);
    id_map.m_count++;
}

/* check & correct map */
/* when last map becomes free allocated MAP_SIZE bytes are freed. */
static void check_map_size()
{
    if (id_map.m_begin == NULL)
        if (id_map.m_used != 0 || id_map.m_count != 0)
            die("check_map_size: corrupted map structure");

    if (id_map.m_begin != NULL)
        if (id_map.m_count == 0)
            die("check_map_size: corrupted map structure");

    if (id_map.m_used * sizeof(__u32) > id_map.m_count * MAP_SIZE)
        die("check_map_size: corrupted map structure");

    if (id_map.m_used * sizeof(__u32) <= (id_map.m_count - 1) * MAP_SIZE)
    {
        if (id_map.m_count != 1)
            id_map.m_begin = expandmem(id_map.m_begin, id_map.m_count * MAP_SIZE, -MAP_SIZE);
        else
            freemem(id_map.m_begin);
        id_map.m_count--;
    }
}

/* free whole map */
void free_id_map()
{
    print_objectid_list ();

    id_map.m_used = 0;
    while(id_map.m_count)
        check_map_size();
}

/* ubin_search_id is used to find id in the map (or proper place to insert the new id) */
/* if smth goes wrong or ubin_search_id stops working properly check_id_search_result
should help to find raised problems */
static int check_id_search_result(int res, __u32 pos, __u32 id)
{
    if (res != ITEM_FOUND && res != ITEM_NOT_FOUND)
        die("check_id_search_result: get wrong result from ubin_search (%d)", res);

    if (res == ITEM_FOUND && *(id_map.m_begin + pos) != id)
        die("check_id_search_result: wrong id found %u %u", id, *(id_map.m_begin + pos));

    if (res == ITEM_NOT_FOUND)
    {
//        if (pos == 0)
//            die("check_id_search_result: id (%u) less than first id in the map %u", id, *(id_map.m_begin));
//        if (pos < 0)
//            die("check_id_search_result: wrong search result %d", pos);
        if (pos > id_map.m_used)
            die("check_id_search_result: get bad position (%u), used %u", pos, id_map.m_used);
        if (pos >= 0 && pos <= id_map.m_used && *(id_map.m_begin + pos - 1) >= id)
            die("check_id_search_result: previous id (%u) not less than (%u)", *(id_map.m_begin + pos - 1), id);
        if (pos >= 0 && pos < id_map.m_used && *(id_map.m_begin + pos) < id)
            die("check_id_search_result: found id (%u) not much than (%u)", *(id_map.m_begin + pos), id);
    }
}

/* return if id is used */
int is_id_used (__u32 id, __u32 * pos)
{
    int i = 0;
    int res;
    __u32 * m_curr;

    *pos = 0;
    if (id_map.m_begin != NULL)
    {/* smth exists in the map, find proper place to insert or this id */
        res = ubin_search_id(&id, id_map.m_begin, id_map.m_used, pos);
        if (opt_verbose)
            check_id_search_result(res, *pos, id);
    }else
        return 0;

    /* pos is either a needed id or a right boundary */
    if (id_map.m_begin == NULL)
        return 0;

    if (res == ITEM_NOT_FOUND) {
	/* if found place to insert is between  (2n-1)th and (2n)th element in the map - this id is not used */
	/* if between (2n)th and (2n+1)th element - used */
        if (fmod(*pos, 2) == 0)
            return 0;
        else
            return 1;
    }else if (res == ITEM_FOUND){
        /* if id was found on (2n) place  in the map - used id */
        /* unused othewise */
        if (fmod(*pos, 2) == 0)
            return 1;
        else
            return 0;
    }
    die("is_id_used: should not get here");
}

/* returns 1 if error occurs, 0 otherwise */
int mark_id_as_used (__u32 id)
{
    __u32 pos;
    __u32 * objectid_map;

    /* check wether id is used and get place if used or place to insert if not */
    if (is_id_used (id, &pos) == 1){
        /*print_objectid_map (s);*/
        return 1;
//        die("mark_id_as_used: id is used already");
    }

    if (fmod(pos, 2) == 0){
        /* ITEM_NOT_FOUND, take a look at is_id_used() */

	  /* no one id in the map */
        if (id_map.m_begin == NULL)
            grow_id_map();

        /* id + 1 is used, change id + 1 to id and exit */
        if ( id + 1 == le32_to_cpu (id_map.m_begin[pos]) ) {
	    id_map.m_begin[pos] = cpu_to_le32 (id);
	    return 0;
        }

        if (id_map.m_count * MAP_SIZE == id_map.m_used * sizeof(__u32))
            grow_id_map();

        if (id_map.m_used - pos > 0)
            memmove (id_map.m_begin + pos + 2, id_map.m_begin + pos, (id_map.m_used - pos) * sizeof (__u32));

        id_map.m_used += 2;
        id_map.m_begin[pos] = cpu_to_le32 (id);
        id_map.m_begin[pos+1] = cpu_to_le32 (id + 1);
        return 0;
    }else if (fmod(pos, 2) == 1){
        /* ITEM_FOUND on (2n+1)th place */
        /* change id to id+1 and exit */
        id_map.m_begin[pos] = cpu_to_le32(id + 1);
        /* if end id of current interval == start id of next interval
           we eliminated a sequence of unused objectids */
        if (id_map.m_begin[pos + 1] == id_map.m_begin[pos]) {
            memmove (id_map.m_begin + pos, id_map.m_begin + pos + 2, (id_map.m_used - pos - 2) * sizeof (__u32));
            id_map.m_used -=2;
            check_map_size();
        }
        return 0;
    }
    die("mark_is_as_used: should not get here");
}

__u32 get_free_id()
{
  /* If map is not NULL return the second element (first position in the map).
     This allocates the first unused objectid. That is to say, the first entry
     on the objectid map is the first unused objectid. */
    if (id_map.m_begin == NULL)
      return 1;
    return (id_map.m_begin[1]);
}

/* print the map of objectids */
void print_objectid_list ()
{
    int i;
    printf ("\n control id map: all %d, used:%d", id_map.m_count * MAP_SIZE, id_map.m_used);

    for (i = 0; i < id_map.m_used; i += 2)
	    printf ("\n[%u-%u]", id_map.m_begin[i], id_map.m_begin[i + 1] - 1);
}

/* print on-disk map of objectids */
void print_disk_objectid_list ()
{
    int i;
    __u32 * objectid_map = (__u32 *)((char *)SB_DISK_SUPER_BLOCK (&g_sb) + (sb_size(&g_sb)));
    printf ("\n on-disk id map. used:%lu", SB_OBJECTID_MAP_SIZE(&g_sb));

    for (i = 0; i < SB_OBJECTID_MAP_SIZE(&g_sb); i += 2)
	    printf ("\n[%u-%u]", objectid_map[i], objectid_map[i + 1] - 1);
}

/*
sinchronize id map to disk for FSCK_CORRECT_BITMAP || FSCK_REBUILD
or check wether on-disk and control maps are the same
*/
void compare_id_maps(struct super_block * s)
{
    __u32 * objectid_map;
    __u32 cur;
    __u32 disk_max, control_max;
    int i;

    objectid_map = (__u32 *)((char *)SB_DISK_SUPER_BLOCK (s) + (sb_size(s)));
    cur = 0;
    control_max = id_map.m_used;

    if (opt_fsck_mode == FSCK_CORRECT_BITMAP || opt_fsck_mode == FSCK_REBUILD) {
      disk_max = SB_OBJECTID_MAP_MAXSIZE(s);
      if (disk_max > control_max) {
        memcpy(objectid_map, id_map.m_begin, control_max * sizeof(__u32));
        memset(objectid_map + control_max, 0, (disk_max - control_max) * sizeof(__u32));
      }else
        memcpy(objectid_map, id_map.m_begin, disk_max * sizeof(__u32));

      set_objectid_map_size ((struct reiserfs_super_block *)(SB_BUFFER_WITH_SB (s)->b_data), id_map.m_used);
      return;
    }

    printf("Comparing objectid maps ..");
    disk_max = SB_OBJECTID_MAP_SIZE(s);

    if (disk_max % 2)   /* must be even */
      die ("compare_id_maps: wrong on-disk map size\n");

    if ( memcmp(objectid_map, id_map.m_begin, sizeof(__u32) *
                  ((disk_max > control_max) ? control_max : disk_max)) ) {
      printf ("\non-disk id map does not match to the correct one");
      return;
    }

    printf("ok\n");
}


