/* 
 * Copyright 2000 by Hans Reiser, licensing governed by reiserfs/README
 */
  
/*
 * 2000/10/26 - Initial version.
 */

#include <assert.h>
#include <misc.h>
#include <reiserfs_lib.h>

/* create clean bitmap */
reiserfs_bitmap_t * reiserfs_create_bitmap (unsigned int bit_count)
{
    reiserfs_bitmap_t * bm;

    bm = getmem (sizeof (reiserfs_bitmap_t));
    if (!bm)
	return 0;
    bm->bm_size = bit_count;
    bm->bm_map = getmem ((bit_count + 7) / 8);
    if (!bm->bm_map) {
	freemem (bm);
	return 0;
    }
    return bm;
}

/* Expand existing bitmap.  Return non-zero if can't. */
int reiserfs_expand_bitmap (reiserfs_bitmap_t * bm, unsigned int bit_count)
{
    unsigned int byte_count = ((bit_count + 7) / 8);
    char * new_map;

    new_map = expandmem (bm->bm_map, byte_count);

    if (!new_map) {
	return 1;
    }
    
    bm->bm_map = new_map;
    bm->bm_size = bit_count;
    return 0;
}

/* bitmap destructor */
void reiserfs_delete_bitmap (reiserfs_bitmap_t * bm)
{
    freemem(bm->bm_map);
    bm->bm_map = NULL;		/* to not reuse bitmap handle */
    bm->bm_size = 0;
    freemem(bm);
}

void reiserfs_bitmap_set_bit (reiserfs_bitmap_t * bm, unsigned int bit_number)
{
    assert(bit_number <= bm->bm_size);
    set_bit(bit_number, bm->bm_map);
}

void reiserfs_bitmap_clear_bit (reiserfs_bitmap_t * bm, unsigned int bit_number)
{
    assert(bit_number <= bm->bm_size);
    clear_bit(bit_number, bm->bm_map);
}

int reiserfs_bitmap_test_bit (reiserfs_bitmap_t * bm, unsigned int bit_number)
{
    assert(bit_number <= bm->bm_size);
    return test_bit(bit_number, bm->bm_map);
}

int reiserfs_bitmap_find_zero_bit (reiserfs_bitmap_t * bm, unsigned int * start)
{
    unsigned int  bit_nr = *start;
    assert(*bit_number < bm->bm_size);

    bit_nr = find_next_zero_bit(bm->bm_map, bm->bm_size, *start);

    if (bit_nr >= bm->bm_size) { /* search failed */	
	return 1;
    }

    *start = bit_nr;
    return 0;
}

/* copy reiserfs filesystem bitmap into memory bitmap */
int reiserfs_fetch_bitmap_from_disk (reiserfs_bitmap_t * bm, reiserfs_filsys_t * fs)
{
    /* reiserfs on-disk format detection code is not written yet, this function depends on it. */
    return 0;
}
